import { world, system, BlockTypes } from "@minecraft/server";
import {ModalFormData} from "@minecraft/server-ui";

const activePedestals = new Map();
const playerCooldowns = new Map();
const COOLDOWN_DURATION = 500; // 500 ms cooldown

const stairs = ["_stairs"];
const blockItemOverrides = ["bed", "sniffer", "ladder", "lever", "button", "cake", "torch", "lantern", "frog_spawn", "chain", "bell", "frame", "brewing_stand", "campfire", "sapling", "deadbush", "hanging", "web", "amethyst", "_candle", "lichen", "vein", "flower", "bars", "clump", "_tulip", "mushroom", "lilac", "vines", "warped_roots", "crimson_roots", "rose", "dandelion", "propagule", "pitcher", "poppy", "bluet", "coral", "orchid", "wart", "fern", "sprout", "grass", "oxeye", "allium", "lily", "petals", "peony", "eyeblossom"];
const stickLikeEndWith = ["stick", "blaze_rod", "breeze_rod", , "fishing_rod", "ender_sceptre"];
const restictedItems = ["_leggings", "_chestplate", "_helmet", "_boots", "shield", "bow", "elytra", "spyglass"]

world.beforeEvents.playerInteractWithBlock.subscribe((e) => {
    const player = e.player;
    const block = e.block;
    const blockLocation = block.location;
    const dimension = player.dimension;
    
    if (block.typeId !== "md_wardens:item_pedestal_block") return;
    e.cancel = true;
    const currentTime = Date.now();
    const lastInteractionTime = playerCooldowns.get(player.id) || 0;
    const itemInHand = player.getComponent("inventory").container.getItem(player.selectedSlotIndex);

    for (let i = 0; i < restictedItems.length; i++) {
        if (itemInHand?.typeId.includes(restictedItems[i])) {
          return;
        }
      }

    if (currentTime - lastInteractionTime < COOLDOWN_DURATION) {
        return; // Still in cooldown period
    }
    playerCooldowns.set(player.id, currentTime);
    const existingEntity = dimension.getEntities({
        location: {x: blockLocation.x, y: blockLocation.y +1, z: blockLocation.z},
        maxDistance: 1,
        type: "md_wardens:item_pedestal"
    })[0];
    if (player.isSneaking) {
      const form = new ModalFormData()
        .title("Pedestal Naming")
        .textField("Enter a name to be applied to the pedestal:", "e.g., 5 Diamonds/Stack")
        .toggle("Lock Pedestal")
      system.runTimeout(() => {
        form.show(player).then((response) => {
          if (response.canceled) return;

          const Name = response.formValues[0]
          const Lock = response.formValues[1]
        if (existingEntity) {
            if (Name.length > 0) {
                existingEntity.nameTag = Name;
            }
            if (Lock) {
                existingEntity.setDynamicProperty("locked", true)
            } else {
                existingEntity.setDynamicProperty("locked", false)
            }
            return;
        } else {
            player.onScreenDisplay.setActionBar("§cPlace an item on the pedestal before naming");
            return;
        }
        });
      }, 1);
    }

    const isLocked = existingEntity?.getDynamicProperty("locked") || false;
    if (isLocked) return;

    system.runTimeout(() => {
        if (existingEntity && !player.isSneaking) {
            const inventory = existingEntity.getComponent("inventory").container;
            const storedItem = inventory.getItem(0);
            
            if (storedItem) {
                existingEntity.triggerEvent('despawn')
                activePedestals.delete(existingEntity.id);
            }
        } else if (itemInHand && !player.isSneaking) {
            const pedestalEntity = dimension.spawnEntity("md_wardens:item_pedestal", {
                x: blockLocation.x + 0.5,
                y: blockLocation.y +1,
                z: blockLocation.z + 0.5,
            });
        
            // Clone the item in hand
            const clonedItem = itemInHand.clone();
            clonedItem.amount = 1;
        
            // Store the cloned item in the entity's inventory
            const inventory = pedestalEntity.getComponent("inventory").container;
            inventory.setItem(0, clonedItem);
        
            // Use replaceitem command to set the visual representation
            pedestalEntity.runCommand(`replaceitem entity @s slot.weapon.mainhand 0 ${itemInHand.typeId} 1`);

            const itemType = getModeId(itemInHand);
            let itemMode = null;
            if (itemType !== undefined) {
                itemMode = itemType+"_equip";
            } else {
                itemMode = "item_equip";
            }
            pedestalEntity.triggerEvent(itemMode);            
        
            // Remove item from player's hand
            if (player.getGameMode() !== "creative") {
                const inv = player.getComponent("inventory").container
                if (itemInHand.amount === 1) {
                  inv.setItem(player.selectedSlotIndex, undefined);
                } else {
                    itemInHand.amount -= 1;
                  inv.setItem(player.selectedSlotIndex, itemInHand);
                }
              }
            activePedestals.set(pedestalEntity.id, blockLocation);
            }
    }, 1);
});

world.beforeEvents.worldInitialize.subscribe(eventData => {
    eventData.blockComponentRegistry.registerCustomComponent('md_wardens:breakPedestal', {
        onPlayerDestroy(e) { 
            const {block, dimension} = e;
            const existingEntity = dimension.getEntities({
                location: {x: block.location.x, y: block.location.y +1, z: block.location.z},
                maxDistance: 1,
                type: "md_wardens:item_pedestal"
            })[0];
            if (existingEntity) {
                existingEntity.triggerEvent("despawn")
            }
        }
    });
});

function getModeId(itemInHand) {
  if (itemInHand.hasTag("minecraft:is_shovel") ||
    itemInHand.hasTag("minecraft:is_hoe") ||
    itemInHand.hasTag("minecraft:is_axe") ||
    itemInHand.hasTag("minecraft:is_pickaxe") ||
    itemInHand.hasTag("minecraft:is_sword")) {
    return "hand";
  }
  if ((itemInHand.typeId.includes("door") && !itemInHand.typeId.includes("trap"))) {
    return "door";
  }
  if(itemInHand.typeId.endsWith("_skull")) {
    return "skull";
  }
  if(itemInHand.typeId === "minecraft:heavy_core") {
    return "core";
  }
  if(itemInHand.typeId.endsWith("_wall") && itemInHand.typeId.includes("minecraft:")) {
    return "wall";
  }
  if(itemInHand.typeId === "minecraft:banner") {
    return "banner";
  }
  for (let i = 0; i < blockItemOverrides.length; i++) {
    if (itemInHand.typeId.includes(blockItemOverrides[i]) && !itemInHand.typeId.includes("azalea") && !itemInHand.typeId.includes("_block") && !itemInHand.typeId.includes("chorus") && !itemInHand.typeId.includes("_stem") && !itemInHand.typeId.includes("_glass") && !itemInHand.typeId.includes("sea_lantern")) {
      return "item";
    }
  }
  for (let i = 0; i < stickLikeEndWith.length; i++) {
    if (itemInHand.typeId.includes(stickLikeEndWith[i])) {
      return "stick";
    }
  }
  try {
    for (let i = 0; i < stairs.length; i++) {
      if (itemInHand.typeId.endsWith(stairs[i])) {
        return "stair";
      }
    }
    const isBlock = BlockTypes.get(itemInHand.typeId)
    if (isBlock) {
        return "block";
    }
  } catch (e) {
    return "item";
  }  
}