import { world, system, ItemStack } from "@minecraft/server";
import { randomTickComponent} from "./OnTick";
import { wardenSet } from "./projectiles"
import "./relics";
import "./projectiles";
import "./pedestalEntity";
import "./itemPedestal";

const sapSet = new Set([
  "md_wardens:bloombelle"
]);

world.afterEvents.entityHealthChanged.subscribe((e) => {
  const {entity, newValue, oldValue} = e;
  if (!wardenSet.has(entity.typeId)) return;
  
  const variantComponent = entity.getComponent("minecraft:variant");
  if (!variantComponent || (variantComponent.value !== 2 && variantComponent.value !== 15)) return;
  let Mob = null;
  if ((variantComponent.value === 15)) {
    Mob = "md_wardens:swampling"
    if (oldValue > 200 && newValue <= 200) {
      summonMobs(entity.location, entity.dimension, 3, Mob);
    }
    
    if (oldValue > 100 && newValue <= 100) {
      summonMobs(entity.location, entity.dimension, 5, Mob);
    }
  
    if (oldValue > 50 && newValue <= 50) {
      summonMobs(entity.location, entity.dimension, 7, Mob);
    }
  }else {
    Mob = "minecraft:husk"
    if (oldValue > 200 && newValue <= 200) {
      summonMobs(entity.location, entity.dimension, 5, Mob);
    }
    
    if (oldValue > 100 && newValue <= 100) {
      summonMobs(entity.location, entity.dimension, 7, Mob);
    }
  
    if (oldValue > 50 && newValue <= 50) {
      summonMobs(entity.location, entity.dimension, 15, Mob);
    }
  }
});

world.beforeEvents.worldInitialize.subscribe(({ blockComponentRegistry }) => {
    blockComponentRegistry.registerCustomComponent("maca:bonemeal", bonemealComponent);
    blockComponentRegistry.registerCustomComponent("maca:random_tick", randomTickComponent);
});

const bonemealComponent = {
    onPlayerInteract(eventData) {
      const { player, block, dimension, face } = eventData;
      const equipment = player.getComponent("equippable");
      const selectedItem = equipment.getEquipment("Mainhand");
  
      if (!selectedItem) return;
  
      if (selectedItem.typeId === "minecraft:bone_meal" && sapSet.has(block.typeId)) {
        const currentStage = getGrowthStage(block);
        if (currentStage === 3) {
              //when max stage is reached I am adding the random tick, then we'll do something over there but could do something here when max stage is reached too
              return;
        } else if (Math.random() < 0.45) {
            const newStage = Math.min(currentStage + 1, 3);
            applyBonemeal(block, dimension, player, equipment, selectedItem, newStage);              
            
          const particleOffset = getParticleOffset(face);
          const particleLocation = {
            x: block.location.x + particleOffset.x + 0.5,
            y: block.location.y + particleOffset.y + 0.5,
            z: block.location.z + particleOffset.z + 0.5
          };
          dimension.spawnParticle("minecraft:crop_growth_emitter", particleLocation);
            }
        } 
    },
  };

function summonMobs(loc, dimension, spawns, Mob) {
  const spawnAttempts = spawns;
  const spawnRadius = 8;

for (let i = 0; i < spawnAttempts; i++) {
    // Generate random offsets within the spawn radius
    const offsetX = (Math.random() * 2 - 1) * spawnRadius;
    const offsetZ = (Math.random() * 2 - 1) * spawnRadius;

    // Calculate the spawn position
    const spawnPos = {
      x: loc.x + offsetX,
      y: loc.y,
      z: loc.z + offsetZ
    };

    // Find the ground level at the spawn position
    const groundY = dimension.getBlockFromRay(spawnPos, { x: 0, y: -1, z: 0 }, { maxDistance: 10 })?.block?.location.y;

  if (groundY !== undefined) {
      spawnPos.y = groundY + 1; // Spawn 1 block above the ground

      // Attempt to spawn the mob
    try {
      dimension.spawnEntity(Mob, spawnPos);
    } catch (error) {
      console.error(`Failed to spawn mob: ${error}`);
    }
  }
}
}

function getParticleOffset(face) {
  switch (face) {
    case "Down": return { x: 0, y: -0.6, z: 0 };
    case "Up": return { x: 0, y: 0.6, z: 0 };
    case "North": return { x: 0, y: 0, z: -0.6 };
    case "South": return { x: 0, y: 0, z: 0.6 };
    case "West": return { x: -0.6, y: 0, z: 0 };
    case "East": return { x: 0.6, y: 0, z: 0 };
    default: return { x: 0, y: 0, z: 0 };
  }
}

function getGrowthStage(block) {
  // Assuming the growth state ID is "maca:growth_stage"
  return block.permutation.getState("maca:growth_stage") || 0;
}

function applyBonemealEffect(
  block,
  dimension,
  player,
  inventory,
  selectedItem
) {
  try {
    dimension.spawnParticle("minecraft:crop_growth_emitter", block.location);
  } catch (error) {
    // Handle or ignore the error
  }
  dimension.playSound("item.bone_meal.use", block.center());

  // Reduce bone meal item
  if (player.getGameMode() !== "creative") {
    if (selectedItem.amount === 1) {
      inventory.setEquipment("Mainhand", undefined);
    } else {
      selectedItem.amount -= 1;
      inventory.setEquipment("Mainhand", selectedItem);
    }
  }
}

function applyBonemeal(
  block,
  dimension,
  player,
  equipment,
  selectedItem,
  newStage
) {
  applyBonemealEffect(block, dimension, player, equipment, selectedItem);

  // Set the new growth stage
  block.setPermutation(
    block.permutation.withState("maca:growth_stage", newStage)
  );
}