import { world, system, ItemStack } from "@minecraft/server";

const BiomeMap = new Map ([
    ["md_wardens:warden_cherry", "md_wardens:cherry_blossom_warden"],
    ["md_wardens:warden_desert", "md_wardens:desert_warden"],
    ["md_wardens:warden_ender", "md_wardens:ender_warden"],
    ["md_wardens:warden_flower", "md_wardens:flower_warden"],
    ["md_wardens:warden_forest", "md_wardens:forest_warden"],
    ["md_wardens:warden_mountain", "md_wardens:mountain_warden"],
    ["md_wardens:warden_savanna", "md_wardens:savanna_warden"],
    ["md_wardens:warden_snowy", "md_wardens:snowy_warden"],
    ["md_wardens:warden_jungle", "md_wardens:jungle_warden"],
    ["md_wardens:warden_ice", "md_wardens:ice_warden"],
    ["md_wardens:warden_swamp", "md_wardens:swamp_warden"],
    ["md_wardens:warden_mesa", "md_wardens:mesa_warden"],
    ["md_wardens:warden_warm_ocean", "md_wardens:warm_ocean_warden"],
    ["md_wardens:warden_mushroom", "md_wardens:mushroom_warden"],
    ["md_wardens:warden_nether", "md_wardens:nether_warden"]
]);

const spawnMap = new Map ([
    ["md_wardens:scorched_dune_shard", "md_wardens:desert_warden"],
    ["md_wardens:woodland_spirit_shard", "md_wardens:forest_warden"],
    ["md_wardens:prismatic_bloom_shard", "md_wardens:flower_warden"],
    ["md_wardens:blossom_essence_shard", "md_wardens:cherry_blossom_warden"],
    ["md_wardens:peak_resonance_shard", "md_wardens:mountain_warden"],
    ["md_wardens:golden_grassland_shard", "md_wardens:savanna_warden"],
    ["md_wardens:glacial_prism_shard", "md_wardens:snowy_warden"],
    ["md_wardens:verdant_pulse_shard", "md_wardens:jungle_warden"],
    ["md_wardens:frost_crystal_shard", "md_wardens:ice_warden"],
    ["md_wardens:murky_essence_shard", "md_wardens:swamp_warden"],
    ["md_wardens:terracotta_essence_shard", "md_wardens:mesa_warden"],
    ["md_wardens:abyssal_tide_shard", "md_wardens:warm_ocean_warden"],
    ["md_wardens:mycelial_fragment_shard", "md_wardens:mushroom_warden"],
    ["md_wardens:blazing_ember_shard", "md_wardens:nether_warden"],
    ["md_wardens:void_resonance_shard", "md_wardens:ender_warden"]
]);

const shardSet = new Set ([
    "md_wardens:blazing_ember_shard",
    "md_wardens:abyssal_tide_shard",
    "md_wardens:golden_grassland_shard",
    "md_wardens:void_resonance_shard",
    "md_wardens:blossom_essence_shard",
    "md_wardens:scorched_dune_shard",
    "md_wardens:glacial_prism_shard",
    "md_wardens:frost_crystal_shard",
    "md_wardens:peak_resonance_shard",
    "md_wardens:verdant_pulse_shard",
    "md_wardens:woodland_spirit_shard",
    "md_wardens:prismatic_bloom_shard",
    "md_wardens:terracotta_essence_shard",
    "md_wardens:mycelial_fragment_shard",
    "md_wardens:murky_essence_shard",
])

const activePedestals = new Map();
const playerCooldowns = new Map();
const COOLDOWN_DURATION = 500; // 500 ms cooldown

world.beforeEvents.playerInteractWithBlock.subscribe((e) => {
    const player = e.player;
    const block = e.block;
    const blockLocation = block.location;
    const dimension = player.dimension;

    if (block.typeId !== "md_wardens:pedestal_block") {
        return;
    }

    if (player.isSneaking) return;
    e.cancel = true;

    const currentTime = Date.now();
    const lastInteractionTime = playerCooldowns.get(player.id) || 0;

    if (currentTime - lastInteractionTime < COOLDOWN_DURATION) {
        return; // Still in cooldown period
    }

    playerCooldowns.set(player.id, currentTime);

    const itemInHand = player.getComponent("inventory").container.getItem(player.selectedSlotIndex);

    system.runTimeout(() => {
        const existingEntity = dimension.getEntities({
            location: blockLocation,
            maxDistance: 1,
            type: "md_wardens:pedestal"
        })[0];

        if (existingEntity) {
            const inventory = existingEntity.getComponent("inventory").container;
            const storedItem = inventory.getItem(0);
            
            if (shardSet.has(storedItem.typeId)) {
                const playerDirection = {
                    x: player.location.x - block.location.x,
                    z: player.location.z - block.location.z
                };
                const playerDirectionLength = Math.sqrt(playerDirection.x * playerDirection.x + playerDirection.z * playerDirection.z);
                const normalizedDir = {
                    x: playerDirection.x / playerDirectionLength,
                    z: playerDirection.z / playerDirectionLength
                };
                const spawnLoc = {x: normalizedDir.x + block.location.x, y: player.location.y, z: normalizedDir.z + block.location.z}
                spawnWarden(existingEntity, player, spawnLoc)
                return;
            }
            
            if (storedItem) {
                existingEntity.triggerEvent('despawn')
                activePedestals.delete(existingEntity.id);
            }
        } else if (shardSet.has(itemInHand?.typeId)) {
            const pedestalEntity = dimension.spawnEntity("md_wardens:pedestal", {
                x: blockLocation.x + 0.5,
                y: blockLocation.y - 0.3,
                z: blockLocation.z + 0.5,
            });
            const validWardens = [];
            for (const [propertyName, wardenType] of BiomeMap) {
            if (pedestalEntity.getProperty(propertyName)) {
                validWardens.push(wardenType);
            }
            }
        
            // Clone the item in hand
            const clonedItem = itemInHand.clone();
        
            // Store the cloned item in the entity's inventory
            const inventory = pedestalEntity.getComponent("inventory").container;
            inventory.setItem(0, clonedItem);
        
            // Use replaceitem command to set the visual representation
            pedestalEntity.runCommand(`replaceitem entity @s slot.weapon.mainhand 0 ${itemInHand.typeId} 1`);
        
            if (itemInHand.hasTag("minecraft:is_shovel") ||
                itemInHand.hasTag("minecraft:is_hoe") ||
                itemInHand.hasTag("minecraft:is_axe") ||
                itemInHand.hasTag("minecraft:is_pickaxe") ||
                itemInHand.hasTag("minecraft:is_sword")) {
                pedestalEntity.triggerEvent('hand_equip');
            } else {
                pedestalEntity.triggerEvent('item_equip');
            }
        
            // Remove item from player's hand
            player.getComponent("inventory").container.setItem(player.selectedSlotIndex, 
            itemInHand.amount > 1 ? new ItemStack(itemInHand.typeId, itemInHand.amount - 1) : undefined);
        
            activePedestals.set(pedestalEntity.id, blockLocation);
            }
    }, 1);
});

function spawnWarden(entity, player, location) {
    const inventory = entity.getComponent("inventory");
    const itemStack = inventory.container.getItem(0);

    const validWardens = [];
    for (const [propertyName, wardenType] of BiomeMap) {
      if (entity.getProperty(propertyName)) {
        validWardens.push(wardenType);
      }
    }
  
    if (validWardens.length === 0) {
      entity.triggerEvent('despawn');
      activePedestals.delete(entity.id);
      player.onScreenDisplay.setActionBar(`Not a Valid Biome`);
      return;
    }
    const entityToSpawn = spawnMap.get(itemStack.typeId);
    if (validWardens.includes(entityToSpawn)) {
        const playerInRange = player.dimension.getEntities({ location: location, maxDistance: 20, type: "player" });
        const wardenSplit = entityToSpawn.split(":")[1].split("_").join(" ");
        playerInRange.forEach(player => {
            player.onScreenDisplay.setTitle(`§3Prepare For Battle`, {
                subtitle: `With the ${wardenSplit.toUpperCase()}`,
                fadeInDuration: 10,
                fadeOutDuration: 10,
                stayDuration: 20 * 5
                });
        });
        entity.triggerEvent('activate');
        player.dimension.spawnParticle("md_wardens:shard_activate", {x: entity.location.x, y: entity.location.y + 1, z: entity.location.z});
        player.dimension.playSound("activate_altar_2", player.location);
        system.runTimeout(() => {
        player.dimension.playSound("activate_altar", player.location);
        }, 4);
        activePedestals.delete(entity.id);
        system.runTimeout(() => {
        player.dimension.spawnEntity(entityToSpawn, location);
        playerInRange.forEach(player => {
            player.addEffect("darkness", 240);
        });
        }, 30);
    } else {
        entity.triggerEvent('despawn');
        activePedestals.delete(entity.id);
        player.onScreenDisplay.setActionBar(`Incorrect biome for the ${itemStack.typeId.split(":")[1].split("_").join(" ")}`);
        system.runTimeout(() => {
        player.onScreenDisplay.setActionBar(`This biome is intended to spawn: ${validWardens.map(w => w.split(":")[1].split("_").join(" ")).join(", ")}`);
        }, 50);
    }
}