import {
  world,
  system,
  ItemStack,
  ItemLockMode,
  EquipmentSlot,
  EntityComponentTypes,
  Player,
  StructureAnimationMode,
  MolangVariableMap,
  BlockPermutation
} from "@minecraft/server";
import { ActionFormData } from "@minecraft/server-ui";
import { startSnowStormAOE } from "./projectiles";

const blockId = "md_wardens:cherry_blossom";
const statePlaced = "maca:placed";
const stateRandom = "maca:random";
const throwPearlID = new Map ();
let fireResistInterval = null;
let conduitInterval = null;
//settings for trail particles on the fireball
const PARTICLE_SPAWN_INTERVAL = 2; // Spawn particles every 2 ticks
const MAX_PARTICLE_DURATION = 100; // Stop spawning particles after 200 ticks (10 seconds)
const PARTICLE_OFFSET = 0.5; // Half a block offset
const cropSet = new Set([
  "minecraft:wheat",
  "minecraft:beetroot",
  "minecraft:carrots",
  "minecraft:potatoes",
]);
const sapSet = new Set ([
  "minecraft:acacia_sapling",
  "minecraft:dark_oak_sapling",
  "minecraft:spruce_sapling",
  "minecraft:oak_sapling",
  "minecraft:birch_sapling",
  "minecraft:mangrove_propagule",
  "minecraft:cherry_sapling",
  "minecraft:jungle_sapling",
  "minecraft:azalea",
  "minecraft:flowering_azalea",
  "minecraft:warped_fungus",
  "minecraft:crimson_fungus",
]);
const dissallowList = new Set([
  "md_wardens:desert_warden",
  "md_wardens:forest_warden",
  "md_wardens:flower_warden",
  "md_wardens:cherry_blossom_warden",
  "md_wardens:mountain_warden",
  "md_wardens:savanna_warden",
  "md_wardens:snowy_warden",
  "md_wardens:jungle_warden",
  "md_wardens:ice_warden",
  "md_wardens:swamp_warden",
  "md_wardens:mesa_warden",
  "md_wardens:warm_ocean_warden",
  "md_wardens:mushroom_warden",
  "md_wardens:nether_warden",
  "md_wardens:ender_warden",
  "minecraft:ender_dragon",
  "minecraft:wither",
  "minecraft:warden",
  "minecraft:ravager",
  "minecraft:ghast",
  "minecraft:item",
  "minecraft:xp_orb",
  "minecraft:player",
  "md_wardens:pedestal",
]);

world.beforeEvents.worldInitialize.subscribe(({ itemComponentRegistry }) => {
  itemComponentRegistry.registerCustomComponent(
    "maca:relics",
    relicsUseComponent
  );
  itemComponentRegistry.registerCustomComponent(
    "maca:relics_useOn",
    relicsUseOnComponent
  );
  itemComponentRegistry.registerCustomComponent(
    "maca:effects",
    effectComponent
  );
  itemComponentRegistry.registerCustomComponent("maca:hitEntity", hitEntity);
  itemComponentRegistry.registerCustomComponent("maca:pearlComponent", pearlComponent);
});

world.beforeEvents.worldInitialize.subscribe(({ blockComponentRegistry }) => {
  blockComponentRegistry.registerCustomComponent("maca:randomTick", randomTick);
  blockComponentRegistry.registerCustomComponent("maca:placeBlossom", placeBlossom);
});

world.beforeEvents.playerInteractWithEntity.subscribe((e) => {
  const entity = e.target;
  const itemStack = e.itemStack;
  const dim = entity.dimension;
  const player = e.player;
  const equippableComponent = player.getComponent(
    EntityComponentTypes.Equippable
  );

  if (itemStack && itemStack.typeId === "md_wardens:ent_stick") {
    if (dissallowList.has(entity.typeId)) return;
    e.cancel = true;
    const entityLocation = entity.location;

    system.runTimeout(() => {
      const nearbyMobs = dim.getEntities({
        location: entityLocation,
        maxDistance: 3,
      });
      nearbyMobs.forEach((nearbyentity) => {
        if (
          nearbyentity.typeId === "minecraft:xp_orb" ||
          nearbyentity.typeId === "minecraft:item"
        ) {
          tempTP(nearbyentity);
        } else if (
          nearbyentity.id !== entity.id &&
          nearbyentity.typeId !== "minecraft:player" &&
          nearbyentity.typeId !== "minecraft:xp_orb" &&
          nearbyentity.typeId !== "minecraft:item" ||
          dissallowList.has(nearbyentity.typeId)
        ) {
          applyKnockback(nearbyentity, entityLocation, 2, 0.5);
          player.dimension.playSound("ent_stick", player.location);
        }
      });
    }, 1);

    // Delay saving the structure by 4 ticks
    system.runTimeout(() => {
      const structureManager = world.structureManager;

      // Define the structure bounds
      const from = {
        x: Math.floor(entityLocation.x),
        y: Math.floor(entityLocation.y) - 1,
        z: Math.floor(entityLocation.z),
      };
      const to = {
        x: Math.floor(entityLocation.x),
        y: Math.floor(entityLocation.y) + 2,
        z: Math.floor(entityLocation.z),
      };

      // Create a unique identifier for the structure
      const identifier = `md_wardens:entity_structure_${entity.id}`;

      try {
        const structure = structureManager.createFromWorld(
          identifier,
          dim,
          from,
          to,
          { includeEntities: true, includeBlocks: false }
        );
        player.dimension.playSound("ent_stick", player.location);

        if (structure) {
          let newItem = new ItemStack("md_wardens:ent_stick_active");
          equippableComponent.setEquipment(EquipmentSlot.Mainhand, newItem);
          const newSelectedItem = equippableComponent.getEquipment(
            EquipmentSlot.Mainhand
          );
          newSelectedItem.setLore([entity.id]);
          equippableComponent.setEquipment(
            EquipmentSlot.Mainhand,
            newSelectedItem
          );
          entity.remove();
        }
      } catch (error) {
        console.error(`Failed to create structure: ${error}`);
      }
    }, 5);
  }
});

world.afterEvents.projectileHitBlock.subscribe((e) => {
  if (e.projectile.typeId !== "md_wardens:suspicious_eye_cluster_projectile")
    return;
  const nearbyPlayers = e.dimension.getEntities({
    location: e.location,
    maxDistance: 60,
    type: "minecraft:player"
  });
  const player = nearbyPlayers.find((player) => throwPearlID.has(player.id));
  if (player) {
    player.teleport(e.location);
    throwPearlID.delete(player.id);
  }
});

world.afterEvents.playerJoin.subscribe(() => {
  system.runTimeout(() => {
    for (const player of world.getAllPlayers()) {
      const heartActive = player.getDynamicProperty("heartActive") || false;
      const conduitActive = player.getDynamicProperty("conduitActive") || false;
      if (heartActive) {
        giveEffect(player, "fire_resistance");
        startfireResistInterval(player);
      }
      if (conduitActive) {
        giveEffect(player, "conduit_power");
        startconduitInterval(player);
      }
    }
  }, 200);
});

world.afterEvents.projectileHitEntity.subscribe((e) => {
  const { dimension, projectile } = e;
  if (projectile.typeId !== "md_wardens:frozen_heart_projectile") return;
  const hitEntity = e.getEntityHit().entity;
  if (
    dissallowList.has(hitEntity.typeId) ||
    hitEntity.typeId === "minecraft:item" ||
    hitEntity.typeId === "minecraft:xp_orb"
  )
    return;
  //summon ice block entity at the location
  dimension.spawnEntity("md_wardens:ice_entity", hitEntity.location);
  //hold the entity up with knockback so they cant move
  let hoverTicks = 0;
  const hoverInterval = system.runInterval(() => {
    if (hoverTicks < 90) {
      try {
        hitEntity.applyKnockback(0, 0, 0, 0.04);
        hoverTicks++;
      } catch (error) {
        system.clearRun(hoverInterval);
      }
    } else if (hoverTicks === 90) {
      system.clearRun(hoverInterval);
    }
  }, 1);
});

world.afterEvents.entityHitBlock.subscribe((e) => {
  const { hitBlock, damagingEntity: player } = e;
  if (!(player instanceof Player)) return;
  const inv = player.getComponent(EntityComponentTypes.Equippable);
  const selectedItem = inv.getEquipment(EquipmentSlot.Mainhand);

  if (
    !selectedItem ||
    (selectedItem.typeId !== "md_wardens:slimey_stick" &&
      selectedItem.typeId !== "md_wardens:suspicious_eye_cluster")
  )
    return;

  if (selectedItem.typeId === "md_wardens:slimey_stick") {
    // Existing slimey stick functionality
    const viewDirection = player.getViewDirection();
    const totalKnockback = 4;
    let horizontalStrength =
      totalKnockback *
      Math.sqrt(
        viewDirection.x * viewDirection.x + viewDirection.z * viewDirection.z
      ) *
      totalKnockback;
    let verticalStrength = totalKnockback * Math.abs(viewDirection.y);
    applyKnockback(
      player,
      hitBlock.location,
      horizontalStrength,
      verticalStrength
    );
    player.dimension.playSound("slime_stick", player.location);
    system.runTimeout(() => catchPlayer(player), 4);
  } else {
    const playerFrame = player.getDynamicProperty("end_portal");
    if (hitBlock.typeId === "minecraft:end_portal_frame" && playerFrame) {
      const portalLocation = JSON.parse(playerFrame);
      const playerPortalDistance = Math.sqrt(
        Math.pow(hitBlock.location.x - portalLocation.x, 2) +
        Math.pow(hitBlock.location.z - portalLocation.z, 2)
      );
      if (playerPortalDistance <= 5) {
        removePortal(player, portalLocation);
        player.dimension.playSound("ent_stick", player.location);
      } else {
        player.onScreenDisplay.setActionBar(`This portal frame is too far from your existing portal`);
      }
    } else if (!playerFrame) {
      placePortal(player, hitBlock);
      player.dimension.playSound("block.end_portal.spawn", player.location)
    } else if (hitBlock.typeId !== "minecraft:end_portal_frame" && playerFrame) {
      player.onScreenDisplay.setActionBar(`You already have an existing portal located at ${playerFrame}`);
    }
  }
});

const placeBlossom = {
  onPlace: (e) => {
    const { block } = e;
    block.setPermutation(
      block.permutation.withState(`${statePlaced}`, true) &&
        block.permutation.withState(
          `${stateRandom}`,
          Math.floor(Math.random() * 4)
        )
    );
  },
  beforeOnPlayerPlace: (e) => {
    const { player, block } = e;
    const placed = BlockPermutation.resolve(`${blockId}`)
      .withState(`${stateRandom}`, Math.floor(Math.random() * 4))
      .withState(`${statePlaced}`, true);
    e.cancel = true;

    if (
      (block.isWaterlogged ||
        (block.typeId === "minecraft:water" &&
          block.permutation.getState("liquid_depth") === 0)) &&
      block.above(1).typeId === "minecraft:air"
    ) {
      block.above(1).setPermutation(placed);
    } // depth 1
    else if (
      (block.isWaterlogged || block.typeId === "minecraft:water") &&
      block.above(1).typeId === "minecraft:water" &&
      block.above(2).typeId === "minecraft:air"
    ) {
      block.above(2).setPermutation(placed);
    } // depth 2
    else if (
      (block.isWaterlogged || block.typeId === "minecraft:water") &&
      block.above(1).typeId === "minecraft:water" &&
      block.above(2).typeId === "minecraft:water" &&
      block.above(3).typeId === "minecraft:air"
    ) {
      block.above(3).setPermutation(placed);
    } // depth 3
    else if (
      (block.isWaterlogged || block.typeId === "minecraft:water") &&
      block.above(1).typeId === "minecraft:water" &&
      block.above(2).typeId === "minecraft:water" &&
      block.above(3).typeId === "minecraft:water" &&
      block.above(4).typeId === "minecraft:air"
    ) {
      block.above(4).setPermutation(placed);
    } // depth 4
    else if (
      (block.isWaterlogged || block.typeId === "minecraft:water") &&
      block.above(1).typeId === "minecraft:water" &&
      block.above(2).typeId === "minecraft:water" &&
      block.above(3).typeId === "minecraft:water" &&
      block.above(4).typeId === "minecraft:water" &&
      block.above(5).typeId === "minecraft:air"
    ) {
      block.above(5).setPermutation(placed);
    } // depth 5
    else return;

    const item = player
      .getComponent("inventory")
      .container.getItem(player.selectedSlotIndex);
    const inventory = player.getComponent("inventory").container;
    const amount = item.amount;
    if (amount == 1) {
      inventory.setItem(player.selectedSlotIndex, null);
    } else {
      item.amount -= 1;
      inventory.setItem(player.selectedSlotIndex, item);
    }
  },
};

const hitEntity = {
  onHitEntity(e) {
    const { attackingEntity: player, hadEffect, hitEntity, itemStack } = e;
    if (!hadEffect) return;
    if (itemStack.typeId === "md_wardens:fungal_dagger") {
      hitEntity.addEffect("fatal_poison", 60);
    } else if (itemStack.typeId === "md_wardens:slimey_stick") {
      player.dimension.playSound("slime_stick", player.location);
      applyKnockback(hitEntity, player.location, 4.5, 1);
    }
  },
};

const randomTick = {
  onTick(e) {
    const baseBlock = e.block;
    const dimension = e.dimension;
    if (Math.random() * 10 > 6) return; // 60% chance to cancel the event

    const isCherry = baseBlock.typeId === "md_wardens:cherry_blossom";
    const range = isCherry ? 4 : 6; // 4 for 9x9, 6 for 13x13
    const saplings = [];
    const crops = [];
    const blocksToGrow = [];

    // Loop through the volume
    for (let x = -range; x <= range; x++) {
      for (let y = -1; y <= 1; y++) {
        for (let z = -range; z <= range; z++) {
          const blockPos = {
            x: Math.floor(baseBlock.location.x + x),
            y: Math.floor(baseBlock.location.y + y),
            z: Math.floor(baseBlock.location.z + z)
          };
          const block = dimension.getBlock(blockPos);
          
          // Check if the block is a sapling
          if (sapSet.has(block?.typeId)) {
            saplings.push(block);
          } else if (cropSet.has(block?.typeId)) {
            crops.push(block);
          }
        }
      }
    }

    // Randomly select up to 2 saplings if there are 3 or more
    if (saplings.length >= 3) {
        for (let i = 0; i < 2 && saplings.length > 0; i++) {
          const randomIndex = Math.floor(Math.random() * saplings.length);
          blocksToGrow.push(saplings.splice(randomIndex, 1)[0]);
        }
      }
  
      // Select crops
      if (crops.length > 0) {
        const numCropsToGrow = Math.min(5, crops.length);
        for (let i = 0; i < numCropsToGrow; i++) {
          const randomIndex = Math.floor(Math.random() * crops.length);
          blocksToGrow.push(crops.splice(randomIndex, 1)[0]);
        }
      }

    // Apply bonemeal to selected blocks
    for (const block of blocksToGrow) {
      applyBonemeal(block, dimension);
    }
}};

const relicsUseOnComponent = {
  onUseOn(e) {
    const { itemStack, source: player } = e;
    const lore = itemStack.getLore();
    const dim = player.dimension;
    switch (itemStack.typeId) {
      case "md_wardens:ent_stick_active":
        if (lore) {
          const loc = player.getBlockFromViewDirection();
          const structureManager = world.structureManager;
          const structure = structureManager.get(
            `md_wardens:entity_structure_${lore}`
          );
          if (structure == void 0) return;
          structureManager.place(structure, player.dimension, loc.block, {
            includeEntities: true,
            includeBlocks: false,
          });
          dim.playSound("ent_stick_reverse", player.location);
          system.runTimeout(() => {
            structureManager.delete(structure);
            const equippableComponent = player.getComponent(
              EntityComponentTypes.Equippable
            );
            let newItem = new ItemStack("md_wardens:ent_stick");
            equippableComponent.setEquipment(EquipmentSlot.Mainhand, newItem);
          }, 4);
        }
        break;
    }
  },
};

const relicsUseComponent = {
  onUse(e) {
    const { itemStack, source: player } = e;
    const inventory = player.getComponent("inventory").container;
    const dim = player.dimension;
    let newItem = undefined;
    switch (itemStack?.typeId) {
      case "md_wardens:volcanic_heart":
        if (player.isSneaking) {
          shootFireball(player);
        } else if (!player.isSneaking && fireResistInterval === null) {
          player.dimension.playSound("activate", player.location);
          newItem = new ItemStack("md_wardens:volcanic_heart_active", 1);
          player.setDynamicProperty("heartActive", true);
          newItem.lockMode = ItemLockMode.inventory;
          newItem.keepOnDeath = true;
          inventory.setItem(player.selectedSlotIndex, newItem);
          giveEffect(player, "fire_resistance");
          startfireResistInterval(player);
        }
        break;
      case "md_wardens:volcanic_heart_active":
        if (!fireResistInterval) {
          newItem = new ItemStack("md_wardens:volcanic_heart", 1);
          inventory.setItem(player.selectedSlotIndex, newItem);
          player.dimension.playSound("deactivate", player.location);
          player.setDynamicProperty("heartActive", false);
          player.dimension.playSound("deactivate", player.location);
        } else if (player.isSneaking) {
          shootFireball(player);
        } else if (!player.isSneaking && fireResistInterval !== null) {
          player.dimension.playSound("deactivate", player.location);
          newItem = new ItemStack("md_wardens:volcanic_heart", 1);
          inventory.setItem(player.selectedSlotIndex, newItem);
          player.setDynamicProperty("heartActive", false);
          system.clearRun(fireResistInterval);
          fireResistInterval = null;
        }
        break;
      case "md_wardens:activated_heart_of_the_sea":
        if (conduitInterval === null) {
          newItem = new ItemStack(
            "md_wardens:activated_heart_of_the_sea_active",
            1
          );
          player.dimension.playSound("activate", player.location);
          player.setDynamicProperty("conduitActive", true);
          newItem.lockMode = ItemLockMode.inventory;
          newItem.keepOnDeath = true;
          inventory.setItem(player.selectedSlotIndex, newItem);
          giveEffect(player, "conduit_power");
          startconduitInterval(player);
        }
        break;
      case "md_wardens:activated_heart_of_the_sea_active":
        if (!player.isSneaking && conduitInterval !== null) {
          player.dimension.playSound("deactivate", player.location);
          newItem = new ItemStack("md_wardens:activated_heart_of_the_sea", 1);
          inventory.setItem(player.selectedSlotIndex, newItem);
          player.setDynamicProperty("conduitActive", false);
          system.clearRun(conduitInterval);
          conduitInterval = null;
        }
        break;
      case "md_wardens:yeti_horn":
        dim.playSound("yeti_horn", player.location);
        startSnowStormAOE(player);
        break;
      case "md_wardens:suspicious_eye_cluster":
        newItem = new ItemStack("md_wardens:suspicious_eye_cluster", 1);
        inventory.setItem(player.selectedSlotIndex, newItem);
        throwPearlID.set(player.id)
        break;
      case "md_wardens:mysterious_core":
        player.dimension.playSound("activate_altar_2", player.location);
        luckMenu(player);
        break;
      case "md_wardens:frozen_heart":
        shootFrozenHeart(player);
        break;
      case "md_wardens:poisonous_stinger":
        poisonStinger(player);
        break;
    }
  },
};

const pearlComponent = {
  onUse(e) {
    const { source: player } = e;
    const existingPortal = player.getDynamicProperty("end_portal") || false;
    let newItem = new ItemStack("md_wardens:suspicious_eye_cluster", 1);
    const inventory = player.getComponent("inventory").container;
    if (existingPortal) {
        const portalLocation = JSON.parse(existingPortal);
        const lore = portalLocation
        ? [`End Portal Location: ${portalLocation.x}, ${portalLocation.y}, ${portalLocation.z}`]
        : [];
        newItem.setLore(lore);
        inventory.setItem(player.selectedSlotIndex, newItem);
        throwPearlID.set(player.id)
    } else {
        inventory.setItem(player.selectedSlotIndex, newItem);
        throwPearlID.set(player.id)
    }
  },
};

const effectComponent = {
  onConsume(e) {
    switch (e.itemStack.typeId) {
      case "md_wardens:baobab_fruit":
        e.source.addEffect("strength", 400, { amplifier: 1 });
        e.source.addEffect("absorption", 1200);
        break;
      case "md_wardens:cherry":
        e.source.addEffect("speed", 2400, { amplifier: 1 });
        e.source.addEffect("regeneration", 2400);
        break;
      case "md_wardens:prickly_fruit":
        e.source.addEffect("haste", 12000, { amplifier: 2 });
        e.source.addEffect("night_vision", 12000);
        break;
    }
  },
};

function removePortal(player, location) {
  const dimension = player.dimension;
  for (let x = -4; x <= 4; x++) {
    for (let y = 0; y <= 1; y++) {
      for (let z = -4; z <= 4; z++) {
        const block = dimension.getBlock({
          x: location.x + x,
          y: location.y + y,
          z: location.z + z,
        });
        if (
          block.typeId === "minecraft:end_portal_frame" ||
          block.typeId === "minecraft:end_portal"
        ) {
          block.setType("minecraft:air");
        }
      }
    }
  }
  player.setDynamicProperty("end_portal", undefined);
  updateItemLore(player, undefined);
}

function placePortal(player, hitBlock) {
  const structureName = "end_portal";
  const location = {
    x: hitBlock.location.x - 2,
    y: hitBlock.location.y + 1,
    z: hitBlock.location.z - 2,
  };
  const dimension = player.dimension;
  for (let x = -2; x <= 2; x++) {
    for (let y = 0; y <= 1; y++) {
      for (let z = -2; z <= 2; z++) {
        const block = dimension.getBlock({
          x: location.x + x,
          y: location.y + y,
          z: location.z + z,
        });
        if (block.typeId !== "minecraft:air") {
          player.onScreenDisplay.setActionBar(
            `Not enough space to place the portal`
          );
          return;
        }
      }
    }
  }

  const placeOptions = {
    animationMode: StructureAnimationMode.Blocks,
    animationSeconds: 10,
    integrity: 1,
  };

  world.structureManager.place(
    structureName,
    player.dimension,
    location,
    placeOptions
  );

  player.setDynamicProperty("end_portal", JSON.stringify(location));
  updateItemLore(player, location);
}

function updateItemLore(player, location) {
  const inv = player.getComponent(EntityComponentTypes.Inventory).container;
  const item = inv.getItem(player.selectedSlotIndex);
  if (item && item.typeId === "md_wardens:suspicious_eye_cluster") {
    const lore = location
      ? [`End Portal Location: ${location.x}, ${location.y}, ${location.z}`]
      : [];
    item.setLore(lore);
    inv.setItem(player.selectedSlotIndex, item);
  }
}

function poisonStinger(player) {
  const centerX = player.location.x;
  const centerY = player.location.y + 1.4;
  const centerZ = player.location.z;
  const radius = 3;
  const particleCount = 25;
  const maxDistance = 15;
  const damageAmount = 6;
  const poisonDuration = 100;
  player.dimension.playSound("pines", player.location);

  for (let i = 0; i < particleCount; i++) {
    const angle = (i / particleCount) * 2 * Math.PI;
    const x = centerX + radius * Math.cos(angle);
    const z = centerZ + radius * Math.sin(angle);

    const dirX = (x - centerX) / radius;
    const dirZ = (z - centerZ) / radius;

    const molangVars = new MolangVariableMap();
    molangVars.setFloat("variable.direction_x", dirX);
    molangVars.setFloat("variable.direction_z", dirZ);
    player.dimension.spawnParticle(
      "md_wardens:cactus_spine_particle",
      { x, y: centerY, z },
      molangVars
    );

    const raycastResult = player.dimension.getEntities({
      location: player.location,
      maxDistance: maxDistance,
    });

    raycastResult.forEach((entity, index) => {
      if (entity.id !== player.id && !dissallowList.has(entity.typeId)) {
        const distance = Math.sqrt(
          Math.pow(entity.location.x - x, 2) +
            Math.pow(entity.location.z - z, 2)
        );
        const delay = Math.floor(distance * 1.2);

        system.runTimeout(() => {
          try {
            entity.applyDamage(damageAmount);
            entity.addEffect("fatal_poison", poisonDuration, {
              amplifier: 0,
              showParticles: true,
            });
          } catch (error) {}
        }, delay);
      }
    });
  }
}

function catchPlayer(player) {
  const catchInterval = system.runInterval(() => {
    const velocity = player.getVelocity();
    const rayDirection = normalizeVector(velocity);

    const raycastOptions = {
      maxDistance: 30,
      includeLiquidBlocks: true,
      includePassableBlocks: false,
    };

    const playerLocation = player.location;
    const checkRadius = 3;

    for (let x = -checkRadius; x <= checkRadius; x++) {
      for (let z = -checkRadius; z <= checkRadius; z++) {
        const checkLocation = {
          x: playerLocation.x + x,
          y: playerLocation.y,
          z: playerLocation.z + z,
        };

        const raycastResult = player?.dimension.getBlockFromRay(
          checkLocation,
          rayDirection,
          raycastOptions
        );

        if (raycastResult) {
          const blockLocation = raycastResult.block.location;
          const distanceToBlock = calculateDistance(
            playerLocation,
            blockLocation
          );

          if (distanceToBlock <= 7) {
            player.addEffect("resistance", 15, {
              showParticles: false,
              amplifier: 100,
            });
            if (blockLocation.y <= playerLocation.y && player.isOnGround) {
              system.clearRun(catchInterval);
            }
            return;
          }
        }
      }
    }
  }, 1);
}

function normalizeVector(vector) {
  const magnitude = Math.sqrt(
    vector.x * vector.x + vector.y * vector.y + vector.z * vector.z
  );
  return {
    x: vector.x / magnitude,
    y: vector.y / magnitude,
    z: vector.z / magnitude,
  };
}

function calculateDistance(pos1, pos2) {
  return Math.sqrt(
    Math.pow(pos1.x - pos2.x, 2) +
      Math.pow(pos1.y - pos2.y, 2) +
      Math.pow(pos1.z - pos2.z, 2)
  );
}

function startfireResistInterval(player) {
  fireResistInterval = system.runInterval(() => {
    giveEffect(player, "fire_resistance");
  }, 600);
}

function startconduitInterval(player) {
  conduitInterval = system.runInterval(() => {
    giveEffect(player, "conduit_power");
  }, 600);
}

function shootFrozenHeart(player) {
  let launchVel = 1.5;
  const velocity = player.getViewDirection();
  let headLoc = player.getHeadLocation();

  const projectile = player.dimension.spawnEntity(
    "md_wardens:frozen_heart_projectile",
    {
      x: headLoc.x + velocity.x,
      y: headLoc.y + velocity.y,
      z: headLoc.z + velocity.z,
    }
  );
  projectile.applyImpulse({
    x: velocity.x * launchVel,
    y: velocity.y * launchVel,
    z: velocity.z * launchVel,
  });
  player.playSound("mob.snowgolem.shoot");

  // Start tracking the projectile for particle effects
  trackProjectileForParticles(projectile);
}

function shootFireball(player) {
  let launchVel = 1.5;
  const velocity = player.getViewDirection();
  let headLoc = player.getHeadLocation();

  const projectile = player.dimension.spawnEntity(
    "md_wardens:fireball_projectile",
    {
      x: headLoc.x + velocity.x,
      y: headLoc.y + velocity.y,
      z: headLoc.z + velocity.z,
    }
  );
  projectile.applyImpulse({
    x: velocity.x * launchVel,
    y: velocity.y * launchVel,
    z: velocity.z * launchVel,
  });
  player.playSound("mob.ghast.fireball");

  // Start tracking the projectile for particle effects
  trackProjectileForParticles(projectile);
}

function trackProjectileForParticles(projectile) {
  let tickCount = 0;
  let lastPosition = projectile.location;
  let particleName = null;
  if (projectile.typeId === "md_wardens:frozen_heart_projectile") {
    particleName = "minecraft:critical_hit_emitter";
  } else if (projectile.typeId === "md_wardens:fireball_projectile") {
    particleName = "minecraft:basic_flame_particle";
  }

  const particleTrackingId = system.runInterval(() => {
    tickCount++;

    // Check if the projectile still exists
    if (!projectile.isValid()) {
      system.clearRun(particleTrackingId);
      return;
    }

    // Spawn particles every PARTICLE_SPAWN_INTERVAL ticks
    if (tickCount % PARTICLE_SPAWN_INTERVAL === 0) {
      const currentPosition = projectile.location;

      // Calculate the movement direction
      const dx = currentPosition.x - lastPosition.x;
      const dy = currentPosition.y - lastPosition.y;
      const dz = currentPosition.z - lastPosition.z;

      // Normalize the direction vector
      const length = Math.sqrt(dx * dx + dy * dy + dz * dz);
      const dirX = length > 0 ? dx / length : 0;
      const dirY = length > 0 ? dy / length : 0;
      const dirZ = length > 0 ? dz / length : 0;

      // Calculate the offset position for the particle
      const particlePosition = {
        x: currentPosition.x - dirX * PARTICLE_OFFSET,
        y: currentPosition.y - dirY * PARTICLE_OFFSET,
        z: currentPosition.z - dirZ * PARTICLE_OFFSET,
      };

      // Spawn the particles at the offset position
      projectile.dimension.spawnParticle(particleName, particlePosition);

      // Update last position for next calculation
      lastPosition = currentPosition;
    }

    // Stop tracking after MAX_PARTICLE_DURATION
    if (tickCount >= MAX_PARTICLE_DURATION) {
      system.clearRun(particleTrackingId);
    }
  }, 1); // Run every tick
}

function giveEffect(player, effectType) {
  player.addEffect(effectType, 1200, { showParticles: false });
}

// Function to calculate a normalized direction vector
function calculateDirection(from, to) {
  const dx = to.x - from.x;
  const dz = to.z - from.z;
  const length = Math.sqrt(dx * dx + dz * dz);
  return { x: dx / length, z: dz / length };
}

function tempTP(nearbyentity) {
  const loc = nearbyentity.location;
  nearbyentity.teleport({ x: loc.x, y: loc.y + 100, z: loc.z });
  system.runTimeout(() => {
    nearbyentity.teleport(loc);
  }, 4);
}

// Function to apply knockback to an entity
export function applyKnockback(
  entity,
  sourcePosition,
  horizontalStrength,
  verticalStrength
) {
  const direction = calculateDirection(sourcePosition, entity.location);
  entity.applyKnockback(
    direction.x,
    direction.z,
    horizontalStrength,
    verticalStrength
  );
}

function applyBonemealEffect(block, dimension) {
  try {
    dimension.spawnParticle("minecraft:crop_growth_emitter", block.center());
  } catch (error) {}
  dimension.playSound("item.bone_meal.use", block.center());
}

function applyBonemeal(block, dimension) {
  if (cropSet.has(block.typeId)) {
    let newStage = block.permutation.getState("growth");

    if (newStage < 7) {
      newStage++;
      applyBonemealEffect(block, dimension);

      // Set the new growth stage
      block.setPermutation(block.permutation.withState("growth", newStage));
    }
  } else if (sapSet.has(block.typeId)) {
    // Grow the tree, theres a random chance this block was selected, if it was we will grow the tree
    applyBonemealEffect(block, dimension);
    block.setPermutation(block.permutation.withState("age_bit", true));
  }
}

function getLuckLevel(itemStack) {
  if (!itemStack) return 0;

  const loreComponent = itemStack.getLore();
  if (!loreComponent) return 0;

  for (const line of loreComponent) {
    if (line.startsWith("Luck: ")) {
      const level = parseInt(line.split(": ")[1]);
      return isNaN(level) ? 0 : level;
    }
  }
  return 0;
}

// Function to determine if extra drops should be given
function shouldDoubleDrop(luckLevel) {
  const chance = luckLevel * 10; // Calculate chance based on luck level
  return Math.random() * 100 < chance;
}

// Store item entities found before block break
let itemsBeforeBreak = [];

// Before block break event
world.beforeEvents.playerBreakBlock.subscribe((event) => {
  const player = event.player;
  const dim = player.dimension;

  // Get all item entities around the player before breaking the block
  const queryOptions = {
    type: "minecraft:item",
    location: player.location,
    maxDistance: 6,
  };

  itemsBeforeBreak = dim.getEntities({
    queryOptions,
  });
});

// After block break event
world.afterEvents.playerBreakBlock.subscribe((event) => {
  const block = event.brokenBlockPermutation;
  const blockLocation = block.location;
  const itemStack = event.itemStackAfterBreak;
  const player = event.player;
  const dim = player.dimension;

  // Get the luck level from the tool's lore
  const luckLevel = getLuckLevel(itemStack);

  if (luckLevel > 0 && shouldDoubleDrop(luckLevel)) {
    // Get all item entities around the block location after breaking the block
    const queryOptions = {
      type: "minecraft:item",
      location: blockLocation,
      maxDistance: 5,
    };

    const itemsAfterBreak = dim.getEntities({
      queryOptions,
    });

    // Find new items that were not present before breaking the block
    const newItems = itemsAfterBreak.filter(
      (item) => !itemsBeforeBreak.includes(item)
    );

    // Double the new items
    newItems.forEach((itemEntity) => {
      dim.spawnItem(new ItemStack(block.type.id, 1), itemEntity.location);
    });
  }
});

function luckMenu(player) {
  const inventory = player.getComponent("inventory").container;
  const LuckMenu = new ActionFormData()
    .title("Tool Selection Menu")
    .body(
      "Select the tool you wish to apply Luck to. Each level adds a 10 percent chance to double your drops! Max Level: 5"
    );

  const toolIndices = [];
  for (let i = 0; i < inventory.size; i++) {
    const item = inventory.getItem(i);
    if (
      item &&
      (item.hasTag("minecraft:is_shovel") ||
        item.hasTag("minecraft:is_hoe") ||
        item.hasTag("minecraft:is_axe") ||
        item.hasTag("minecraft:is_pickaxe"))
    ) {
      toolIndices.push(i);
      const luckLvl = getLuckLevel(item);
      if (luckLvl < 5) {
        LuckMenu.button(
          `${
            item.nameTag ? item.nameTag : item.typeId.split(":")[1]
          } - luck lvl: ${luckLvl}`
        );
      }
    }
  }

  if (toolIndices.length === 0) {
    LuckMenu.button("Tools need to be in your Inventory");
  }

  LuckMenu.show(player)
    .then((response) => {
      if (response.canceled) return;

      if (toolIndices.length === 0) {
        player.onScreenDisplay.setActionBar(
          "No eligible tools found in your inventory"
        );
        return;
      }

      const selectedSlot = toolIndices[response.selection];
      const selectedItem = inventory.getItem(selectedSlot);

      if (selectedItem) {
        const newLvl = getLuckLevel(selectedItem) + 1;
        const loreString = `Luck: ${newLvl}`;
        selectedItem.setLore([loreString]);

        inventory.setItem(selectedSlot, selectedItem);
        inventory.setItem(player.selectedSlotIndex, undefined);

        player.onScreenDisplay.setActionBar(
          `Luck applied to ${selectedItem.typeId.split(":")[1]}`
        );
      }
    })
    .catch((error) => {
      console.error("Error setting luck level:", error);
    });
}
